from ds_format.cmd import UsageError
import ds_format as ds
import aquarius_time as aq

def pretty(x, var):
	if var.get('units') == 'days since -4713-11-24 12:00 UTC' and \
	   var.get('calendar') == 'proleptic_gregorian':
		return aq.to_iso(x)
	return str(x)

def cat(vars_, *input_, **opts):
	if not isinstance(vars_, list):
		vars_ = [vars_]
	for filename in input_:
		d = ds.read(filename, vars_,
			full=False,
			jd=(opts.get('jd') or opts.get('h')),
		)
		varsx = [x for x in d.keys() if not x.startswith('.')]
		if len(varsx) == 0:
			return
		dims0 = d['.'][varsx[0]]['.dims']
		for var in varsx:
			dims = d['.'][var]['.dims']
			if dims != dims0:
				raise ValueError('incompatible dimensions')
		n = d[vars_[0]].flatten().shape[0]
		for i in range(n):
			if opts.get('h'):
				s = ','.join([pretty(d[var].flatten()[i], d['.'][var]) for var in vars_])
			else:
				s = ','.join([str(d[var].flatten()[i]) for var in vars_])
			print(s)



import os
import re
from netCDF4 import Dataset
import cftime
import numpy as np
import ds_format as ds
import datetime as dt
import aquarius_time as aq

JD_UNITS = 'days since -4713-11-24 12:00 UTC'
JD_CALENDAR = 'proleptic_greogorian'

READ_EXT = ['nc', 'nc4', 'nc3', 'netcdf', 'hdf', 'h5']
WRITE_EXT = ['nc', 'nc4', 'netcdf']

def detect(filename):
	try:
		f = Dataset(filename)
		return True
	except:
		return False

def read_attrs(var):
	d = {}
	for name in var.ncattrs():
		d[name] = var.getncattr(name)
	return d

def read_var(f, name, sel=None, data=True):
	var = f[name]
	x = None
	attrs = read_attrs(var)
	dims = var.dimensions
	size = var.shape
	if data:
		if sel:
			s = ds.misc.sel_slice(sel, dims)
			x = var[s]
			dims = ds.misc.sel_dims(sel, dims)
		else:
			try:
				x = var[::] if data else None
			except ValueError:
				var.set_auto_mask(False)
				x = var[::] if data else None
	attrs.update({
		'.dims': dims,
		'.size': size,
	})
	return [x, attrs]

def process_datetime_var(d, name):
	if not isinstance(d[name], np.ndarray):
		return
	x = d[name].flatten()
	if len(x) == 0:
		return
	shape = d[name].shape
	units = d['.'][name].get(u'units')
	calendar = d['.'][name].get(u'calendar', u'standard')
	if re.match(r'^days since -4712-01-01[T ]12:00(:00)?( UTC)?$', units) and \
	   calendar in (None, 'standard'):
		units = 'days since -4713-11-24 12:00 UTC'
		calendar = 'proleptic_gregorian'
	try:
		x = cftime.num2date(x, units,
			calendar=calendar,
			only_use_cftime_datetimes=False,
		)
	except: return
	if not (
		isinstance(x[0], cftime.real_datetime) or
		isinstance(x[0], cftime.datetime) or
		isinstance(x[0], dt.datetime)
	):
		return
	if isinstance(x[0], cftime.real_datetime) or \
	   isinstance(x[0], cftime.datetime):
		for i in range(len(x)):
			x[i] = dt.datetime(x[i].year, 1, 1) + \
			(x[i] - type(x[i])(x[i].year, 1, 1))
	d[name] = aq.from_datetime(list(x)).reshape(shape)
	d['.'][name]['units'] = 'days since -4713-11-24 12:00 UTC'
	d['.'][name]['calendar'] = 'proleptic_gregorian'

def read(filename, variables=None, sel=None, full=False, jd=False):
	if type(filename) is bytes and str != bytes:
		filename = os.fsdecode(filename)
	with Dataset(filename, 'r') as f:
		d = {}
		d['.'] = {}
		d['.']['.'] = read_attrs(f)
		for name in f.variables.keys():
			if variables is not None and name not in variables:
				if full:
					_, d['.'][name] = read_var(f, name, sel, False)
			else:
				d[name], d['.'][name] = read_var(f, name, sel)
	if jd:
		for name in ds.get_vars(d):
			process_datetime_var(d, name)
	return d

def write(filename, d):
	ds.validate(d)
	if type(filename) is bytes and hasattr(os, 'fsdecode'):
		filename = os.fsdecode(filename)
	with Dataset(filename, 'w') as f:
		dims = ds.get_dims(d)
		for k, v in dims.items():
			f.createDimension(k, v)
		for name in ds.get_vars(d):
			data = ds.get_var(d, name)
			if data.dtype == 'O' and \
				len(data.flatten()) > 0 and \
				type(data.flatten()[0]) is str:
				dtype = str
			else:
				dtype = data.dtype
			v = f.createVariable(name, dtype, ds.get_dims(d, name))
			v.setncatts(ds.get_attrs(d, name))
			v[::] = data
		f.setncatts(ds.get_attrs(d))

from_netcdf = read
to_netcdf = write



import json
import numpy as np
import ds_format as ds
import datetime as dt
import aquarius_time as aq

READ_EXT = ['json']
WRITE_EXT = ['json']

def read(filename, variables=None, sel=None, full=False, jd=False):
	with open(filename) as f:
		d = json.load(f)
		for var in ds.get_vars(d):
			if type(d[var]) is list:
				d[var] = np.array(d[var])
	return d

def write(filename, d):
	d2 = ds.copy(d)
	for var in ds.get_vars(d2):
		if isinstance(d2[var], np.ndarray):
			d2[var] = d2[var].tolist()
	with open(filename, 'w') as f:
		json.dump(d2, f)

from_json = read
to_json = write



import ds_format as ds
import json
from ds_format.cmd import UsageError, NumpyEncoder

def get(*args, **opts):
	if len(args) != 2:
		raise TypeError('Usage: get <path> <input>')
	path = args[0].split('/')
	input_ = args[1]

	d = ds.from_netcdf(input_, [])

	if len(path) == 2 and path[0] == '' and path[1] == '':
		j = json.dumps(d['.']['.'], sort_keys=True, indent=4, cls=NumpyEncoder)
		print(j)
	elif len(path) == 2 and path[0] == '':
		attr = path[1]
		print(d['.']['.'][attr])
	elif len(path) == 2:
		var = path[0]
		attr = path[1]
		print(d['.'][var][attr])
	elif len(path) == 1:
		var = path[0]
		j = json.dumps(d['.'][var], sort_keys=True, indent=4, cls=NumpyEncoder)
		print(j)



import json
import numpy as np
import ds_format as ds

def ls(*args, **opts):
	input_ = args
	for filename in input_:
		d = ds.read(filename, [], full=True)
		vars_ = sorted([x for x in d['.'].keys() if not x.startswith('.')])
		if opts.get('l'):
			for x in vars_:
				dims = d['.'][x]['.dims']
				size = d['.'][x]['.size']
				dims_s = ','.join([
					'%s=%d' % (dim, size0)
					for dim, size0 in zip(dims, size)
				])
				print('%s(%s)' % (x, dims_s))
		else:
			print('\n'.join(vars_))



from collections import Mapping, Iterable
from .op import *

def validate(d):
	if not isinstance(d, Mapping):
		raise ValueError('dataset must be a mapping')
	if '.' in d and not isinstance(d['.'], Mapping):
		raise ValueError('dataset metadata "." must be a mapping')
	if '.' in d and '.' in d['.'] and not isinstance(d['.']['.'], Mapping):
		raise ValueError('dataset attributes "./." must be a dictionary')
	dim_size = {}
	for name in get_vars(d):
		dims = get_dims(d, name)
		data = get_var(d, name)
		if len(dims) != data.ndim:
			raise ValueError('dataset variable "%s" has inconsistent number of dimensions: %d dimensions defined ("%s") but data has %d dimensions' % (name, len(dims), dims, data.ndim))
		for i, dim in enumerate(dims):
			s = data.shape[i]
			if dim in dim_size and dim_size[dim] != s:
				raise ValueError('dimension "%s" has inconsistent size between variables' % dim)
			dim_size[dim] = s
		if not isinstance(dims, Iterable):
			raise ValueError('".dims" must be iterable')



from .misc import UsageError, NumpyEncoder

from .merge import merge
from .meta import meta
from .ls import ls
from .cat import cat
from .stats import stats
#from .dims import dims
from .get import get
from .select import select
from .write import write

CMDS = {
	'merge': merge,
	'meta': meta,
	'ls': ls,
	'cat': cat,
	'stats': stats,
	'get': get,
	'select': select,
	'write': write,
#	'dims': dims,
}



from collections import Mapping, Iterable
import copy as copy_
import numpy as np
import datetime as dt
from . import misc

def select_var(d, name, sel):
	var_dims = list(d['.'][name]['.dims'])
	d['.'][name]['.dims'] = var_dims
	for key, value in sel.items():
		if isinstance(value, Mapping):
			if len(sel) > 1: raise ValueError('invalid selector')
			newdim = key
			dims = value.keys()
			idxs = value.values()
			selector = tuple([
				idxs[dims.index(var_dim)] if var_dim in dims else slice(None)
				for var_dim in var_dims
			])
			d[name] = d[name][selector]
			for dim in dims:
				if dim in var_dims:
					var_dims.remove(dim)
			d['.'][name]['.dims'].append(newdim)
		else:
			dim, idxs = key, value
			idxs = np.array(idxs) if type(idxs) in (list, tuple) else idxs
			if isinstance(idxs, np.ndarray) and idxs.dtype == np.bool:
				idxs = np.nonzero(idxs)[0]
			if dim in var_dims:
				i = var_dims.index(dim)
				d[name] = np.take(d[name], idxs, axis=i)
				if not isinstance(idxs, np.ndarray):
					var_dims.remove(dim)

def filter_hidden(x):
	if isinstance(x, Mapping):
		return {k: v for k, v in x.items() if not k.startswith('.')}
	if isinstance(x, Iterable):
		return [k for k in x if not k.startswith('.')]
	return x

def select(d, sel):
	for name in d.keys():
		if name.startswith('.'):
			continue
		select_var(d, name, sel)

def get_dims(d, name=None):
	if name is None:
		dims = {}
		for name in get_vars(d):
			data = get_var(d, name)
			for i, dim in enumerate(get_dims(d, name)):
				dims[dim] = data.shape[i]
		return dims
	else:
		try: return d['.'][name]['.dims']
		except KeyError: return gen_dims(d, name)

def get_vars(d):
	return filter_hidden(d.keys())

def get_var(d, name):
	data = d[name]
	if type(data) is np.ndarray:
		return data
	else:
		return np.array(data)

def get_meta(d, name=None):
	if name is None:
		return d.get('.', {})
	else:
		try: return d['.'][name]
		except KeyError: return {}

def get_attrs(d, name=None):
	if name is None:
		try: return filter_hidden(d['.']['.'])
		except KeyError: return {}
	else:
		try: return filter_hidden(d['.'][name])
		except KeyError: return {}

def gen_dims(d, name):
	data = get_var(d, name)
	return [name + ('_%d' % i) for i in range(1, data.ndim + 1)]

def parse_time(t):
	formats = [
		'%Y-%m-%d %H:%M:%S.%f',
		'%Y-%m-%d %H:%M:%S',
		'%Y-%m-%dT%H:%M:%SZ',
	]
	for f in formats:
		try: return dt.datetime.strptime(t, f)
		except: pass
	return None

def time_dt(time):
	return [parse_time(t) for t in time]

def merge_var(dd, var, dim):
	if len(dd) == 0:
		return None, None
	x0 = dd[0][var]
	meta0 = dd[0]['.'][var]
	dims0 = meta0['.dims']
	meta = copy_.deepcopy(meta0)
	if dim in dims0:
		i = dims0.index(dim)
		x = np.concatenate(
			[d[var] for d in dd if d['.'][var]['.dims'] == dims0],
			axis=i
		)
	else:
		meta['.dims'] = [dim] + list(meta['.dims'])
		x = np.stack([d[var] for d in dd if d['.'][var]['.dims'] == dims0])
	return x, meta

def merge(dd, dim, new=False, variables=None):
	dx = {'.': {'.': {}}}
	vars_ = [x for d in dd for x in get_vars(d)]
	dims = [k for d in dd for k in get_dims(d).keys()]
	new = dim not in dims
	for var in vars_:
		if (new and (variables is None or var in variables)) or \
		   dim in get_dims(dd[0], var):
			x, meta = merge_var(dd, var, dim=dim)
		else:
			x, meta = dd[0][var], dd[0]['.'][var]
		dx[var] = x
		dx['.'][var] = meta
	for d in dd:
		if '.' in d['.']:
			dx['.']['.'].update(d['.']['.'])
	return dx

def rename_dim(d, old, new):
	if old == new:
		return
	if '.' in d:
		for var in d['.'].keys():
			meta = d['.'][var]
			if '.dims' in d['.'][var]:
				dims = d['.'][var]['.dims']
				for i, dim in enumerate(dims):
					if dim == old:
						dims[i] = new

def rename(d, old, new):
	if old == new:
		return
	if old in d:
		d[new] = d[old]
		d['.'][new] = d['.'][old]
		del d[old]
		del d['.'][old]
	rename_dim(d, old, new)

def copy(d):
	d2 = {}
	for var in get_vars(d):
		d2[var] = d[var]
	d2['.'] = copy_.deepcopy(d['.'])
	return d2

def group_by(d, dim, group, func):
	groups = sorted(list(set(group)))
	vars = get_vars(d)
	n = len(groups)
	for var in vars:
		dims = d['.'][var]['.dims']
		try:
			i = dims.index(dim)
		except ValueError:
			continue
		size = list(d[var].shape)
		size[i] = n
		x = np.empty(size, d[var].dtype)
		for j, g in enumerate(groups):
			mask = group == g
			slice_x = misc.sel_slice({dim: j}, dims)
			slice_y = misc.sel_slice({dim: mask}, dims)
			y = d[var][slice_y]
			x[slice_x] = func(y, axis=i)
		d[var] = x



import numpy as np

def sel_slice(sel, dims):
	return tuple([
		slice(None) if dim not in sel.keys() else sel[dim]
		for dim in dims
	])

def sel_dims(sel, dims):
	return [
		dim
		for dim in dims
		if dim not in sel.keys() or \
			isinstance(sel[dim], np.ndarray) or \
			type(sel[dim]) in (list, tuple)
	]



import json
import numpy as np
import re
import ds_format as ds

def write(*args, **opts):
	d = {'.': {}}
	output = args[0]
	variables = args[1:]
	for var in variables:
		attrs = None
		if len(var) == 3:
			name, dims, values = var
		elif len(var) == 4:
			name, dims, values, attrs = var
		else:
			raise ValueError('Invalid variable: %s' % var)
		d[name] = np.array(values)
		d['.'][name] = attrs if attrs is not None else {}
		d['.'][name]['.dims'] = dims if isinstance(dims, list) \
			else [dims]
	if len(opts.keys()) > 0:	
		d['.']['.'] = opts
	ds.write(output, d)	



import json
import numpy as np
import ds_format as ds

def meta(*args, **opts):
	input_ = args
	for filename in input_:
		d = ds.read(filename, [], full=True)
		info = d['.']
		j = json.dumps(info, sort_keys=True, indent=4, cls=ds.cmd.NumpyEncoder)
		print(j)



import ds_format as ds

def dims(*args, **opts):
	if len(args) == 0:
		raise TypeError('Usage: dims <input>...')
	input_ = args
	for filename in input_:
		d = ds.read(filename, [])
		dims = ds.get_dims(d)
		for dim in dims:
			print(dim)
		out = {x: for x in dims}
		j = json.dumps(out, indent=4)
		print(j)



import os
import traceback as tb
from .drivers import DRIVERS
import ds_format as ds
import numpy as np

def index(dirname, variables=None, warnings=[], **kwargs):
	l = sorted(os.listdir(dirname))
	dd = []
	for name in l:
		filename = os.path.join(dirname, name)
		try: d = ds.read(filename, variables=variables, **kwargs)
		except Exception as e:
			warnings.append((
				'%s: %s' % (filename, e),
				tb.format_exc()
			))
			continue
		d['filename'] = filename
		d['.']['filename'] = {
			'.dims': [],
		}
		dd.append(d)
	return dd

def read(filename, *args, **kwargs):
	if not os.path.exists(filename):
		raise IOError('%s: File does not exist' % filename)
	for name, driver in DRIVERS.items():
		for ext in driver.READ_EXT:
			end = '.' + ext
			if type(filename) is bytes:
				end = end.encode('utf-8')
			if filename.endswith(end):
				d = driver.read(filename, *args, **kwargs)
				return d
	raise IOError('%s: Unknown file format' % filename)

def readdir(dirname, variables=None, merge=None, warnings=[], **kwargs):
	l = sorted(os.listdir(dirname))
	dd = []
	for name in l:
		filename = os.path.join(dirname, name)
		try: d = ds.read(filename, variables=variables, **kwargs)
		except Exception as e:
			warnings.append((
				'%s: %s' % (filename, e),
				tb.format_exc()
			))
			continue
		d['filename'] = filename
		d['.']['filename'] = {
			'.dims': [],
		}
		dd.append(d)
	if merge is None:
		return dd
	else:
		n = 0
		for n, d in enumerate(dd):
			m = ds.get_dims(d)[merge]
			d['n'] = np.full(m, n)
			d['i'] = np.arange(m)
			d['.']['n'] = {
				'.dims': [merge],
			}
			d['.']['i'] = {
				'.dims': [merge],
			}
		d = ds.op.merge(dd, merge, new='n')
		return d

def write(filename, d):
	for name, driver in DRIVERS.items():
		for ext in driver.WRITE_EXT:
			end = '.' + ext
			if type(filename) is bytes:
				end = end.encode('utf-8')
			if filename.endswith(end):
				driver.write(filename, d)
				return
	raise ValueError('%s: Unknown file extension' % filename)



import ds_format as ds

def merge(dim, *args, **opts):
	input_ = args[:-1]
	output = args[-1]
	dd = []
	for filename in input_:
		d = ds.read(filename)
		dd.append(d)
	d = ds.op.merge(dd, dim,
		variables=opts.get('variables')
	)
	ds.write(output, d)



import numpy as np
import json
import ds_format as ds
from ds_format.cmd import UsageError, NumpyEncoder

def stats(*args, **opts):
	if len(args) < 2:
		raise TypeError('Usage: stats <var> <input>...')
	var = args[0]
	input_ = args[1:]
	for filename in input_:
		d = ds.read(filename, [var])
		x = d[var].flatten()
		count = len(x)
		min_ = np.min(x)
		max_ = np.max(x)
		mean = np.mean(x)
		median = np.median(x)
		j = json.dumps({
			'count': count,
			'min': min_,
			'max': max_,
			'mean': mean,
			'median': median,
		}, cls=NumpyEncoder, sort_keys=True, indent=4)
		print(j)



import ds_format as ds
from ds_format.cmd import UsageError

def select(input_, output, variables=None, sel=None):
	sel = sel[0] if sel is not None and len(sel) > 0 else None
	d = ds.read(input_, variables, sel)
	ds.write(output, d)



import json
import numpy as np

class NumpyEncoder(json.JSONEncoder):
	def default(self, obj):
		if isinstance(obj, np.ndarray):
			return obj.tolist()
		if isinstance(obj, np.generic):
			return np.asscalar(obj)
		return json.JSONEncoder.default(self, obj)

class UsageError(TypeError):
	pass



from .drivers.netcdf import from_netcdf, to_netcdf
from .op import \
	copy, \
	get_attrs, \
	get_dims, \
	get_meta, \
	get_var, \
	get_vars, \
	group_by, \
	merge, \
	rename, \
	rename_dim, \
	select, \
	time_dt
from .io import read, readdir, index, write
from .validate import validate
from . import cmd
from . import misc



#!/usr/bin/env python

from setuptools import setup, find_packages

setup(
    name='ds-format',
    version='1.1.1',
    scripts=['bin/ds'],
    packages=find_packages(),
    description='Python implementation of a dataset format DS for storing data along with metadata',
    author='Peter Kuma',
    author_email='peter@peterkuma.net',
    keywords=['dataset', 'netcdf', 'hdf', 'json', 'numpy'],
    url='https://github.com/peterkuma/ds-format',
    platforms=['any'],
    install_requires=[
		'netCDF4',
		'cftime>=1.5.1',
		'pst-format>=1.1.1',
		'aquarius-time>=0.1.0'
	],
    license='MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
    ],
)
